/* copy_exif - copy EXIF data from one JPEG to another
**
** Copyright (C)2004,2005,2019 by Jef Poskanzer <jef@mail.acme.com>.
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* A few JPEG markers. */
#define JPEG_BEGIN 0xFF
#define JPEG_SOI 0xD8
#define JPEG_APP0 0xE0
#define JPEG_APP1 0xE1
#define JPEG_SOF0 0xC0
#define JPEG_SOF15 0xCF
#define JPEG_SOS 0xDA
#define JPEG_EOI 0xD9

/* And some EXIF tags. */
#define EXIF_ImageWidth 0xa002
#define EXIF_ImageHeight 0xa003


/* Globals. */

static const char* argv0;


/* Forwards. */
static int read_block( FILE* fp, unsigned char** blockP, size_t* block_sizeP, size_t* block_lenP );
static void expand_block( unsigned char** blockP, size_t* block_sizeP, size_t new_size );


int
main( int argc, const char* argv[] )
    {
    const char* source_filename;
    const char* dest_filename;
    FILE* source_fp;
    FILE* dest_fp;
    FILE* temp_fp;
    unsigned char* block;
    size_t block_size, block_len;
    unsigned char* exif_block;
    size_t exif_block_len = 0;
    int width, height;

    argv0 = argv[0];
    if ( argc == 2 )
	{
	source_filename = argv[1];
	dest_filename = "-";
	}
    else if ( argc == 3 )
	{
	source_filename = argv[1];
	dest_filename = argv[2];
	}
    else
	{
	(void) fprintf( stderr, "usage: %s source [dest]\n", argv0 );
	exit( 1 );
        }

    /* JPEG blocks consist of a 0xff, a marker byte, a block size
    ** (two bytes, big-endian), and the rest of the block.  The
    ** block size includes itself - i.e. if the block size was 2 then
    ** there would be no additional bytes in the block.  Yes, this
    ** is stupid.
    **
    ** Regular JPEG files contain an SOI block and then an APP0 block
    ** of type JFIF.  The JFIF block is required to come immediately
    ** after the SOI.
    **
    ** JPEG files with EXIF data contain an SOI block and then an
    ** APP1 block of type EXIF.  The EXIF block is required to come
    ** immediately after the SOI.  Yes, that's a conflict.
    **
    ** The consensus resolution is that files with an EXIF block
    ** should not have a JFIF block at all.  A few programs try to
    ** compromise by putting both blocks in the file.  This code
    ** should be able to read any variety, and will produce files
    ** with just one EXIF block.
    */

    /* Read the source file, looking for the APP1 EXIF block, and save it. */
    source_fp = fopen( source_filename, "r" );
    if ( source_fp == (FILE*) 0 )
	{
	perror( source_filename );
	exit( 1 );
	}
    block = (unsigned char*) 0;
    exif_block = (unsigned char*) 0;
    while ( read_block( source_fp, &block, &block_size, &block_len ) )
	{
	/* Is it an EXIF block? */
	if ( block[1] == JPEG_APP1 && block_len >= 10 &&
	     block[4] == 'E' && block[5] == 'x' && block[6] == 'i' &&
	     block[7] == 'f' && block[8] == 0 && block[9] == 0 )
	    {
	    exif_block = block;
	    exif_block_len = block_len;
	    block = (unsigned char*) 0;
	    break;	/* don't need to read the rest of the file */
	    }
	}
    (void) fclose( source_fp );
    /* If there was no EXIF block to copy, then we're done. */
    if ( exif_block == (unsigned char*) 0 )
	{
	(void) fprintf( stderr, "%s: no EXIF block found in source file\n", argv0 );
	exit( 0 );
	}

    /* Read the destination file, copying blocks to a temp file. */
    if ( strcmp( dest_filename, "-" ) == 0 )
	dest_fp = stdin;
    else
	dest_fp = fopen( dest_filename, "r+" );
    if ( dest_fp == (FILE*) 0 )
	{
	perror( dest_filename );
	exit( 1 );
	}
    temp_fp = tmpfile();
    if ( temp_fp == (FILE*) 0 )
	{
	perror( "opening temporary file" );
	exit( 1 );
	}
    while ( read_block( dest_fp, &block, &block_size, &block_len ) )
	{
	/* Any old JFIF or EXIF blocks in the destination file get dropped. */
	if ( block[1] == JPEG_APP0 && block_len >= 16 &&
	     block[4] == 'J' && block[5] == 'F' && block[6] == 'I' &&
	     block[7] == 'F' && block[8] == 0 )
	    { }
	else if ( block[1] == JPEG_APP1 && block_len >= 10 &&
	     block[4] == 'E' && block[5] == 'x' && block[6] == 'i' &&
	     block[7] == 'f' && block[8] == 0 && block[9] == 0 )
	    { }
	else
	    {
	    if ( fwrite( block, block_len, 1, temp_fp ) != 1 )
		{
		(void) fprintf( stderr, "%s: problem copying to temporary file\n", argv0 );
		exit( 1 );
		}
	    if ( block[1] >= JPEG_SOF0 && block[1] <= JPEG_SOF15 && block_len >= 9 )
		{
		/* The width and height of the destination file get saved.
		** Currently we don't do anything with them, but in a future
		** version we could modify the dimensions in the EXIF block
		** to correspond to the destination file's actual dimensions.
		** The EXIF tags are ExifImageWidth and ExifImageHeight.
		*/
		height = (int) ( block[5] ) * 256 + (int) ( block[6] );
		width = (int) ( block[7] ) * 256 + (int) ( block[8] );
		}
	    }
	}

    /* Copy the temp file back to the destination file. */
    if ( strcmp( dest_filename, "-" ) == 0 )
	dest_fp = stdout;
    else
	rewind( dest_fp );
    rewind( temp_fp );
    while ( read_block( temp_fp, &block, &block_size, &block_len ) )
	{
	/* Write the block. */
	if ( fwrite( block, block_len, 1, dest_fp ) != 1 )
	    {
	    (void) fprintf( stderr, "%s: problem copying back to destination file\n", argv0 );
	    exit( 1 );
	    }
	if ( block[1] == JPEG_SOI )
	    {
	    /* Insert the saved EXIF block into the destination right
	    ** after the SOI.
	    */
	    if ( fwrite( exif_block, exif_block_len, 1, dest_fp ) != 1 )
		{
		(void) fprintf( stderr, "%s: problem copying EXIF block to destination file\n", argv0 );
		exit( 1 );
		}
	    }
	}
    if ( strcmp( dest_filename, "-" ) != 0 )
	(void) fclose( dest_fp );
    (void) fclose( temp_fp );

    exit( 0 );
    }


static int
read_block( FILE* fp, unsigned char** blockP, size_t* block_sizeP, size_t* block_lenP )
    {
    int ich;
    size_t stoopid_len, data_len;
    int got_begin;

    /* Initialize the block if necessary. */
    if ( *blockP == (unsigned char*) 0 )
	{
	*block_sizeP = 10000;		/* arbitrary */
	*blockP = (unsigned char*) malloc( *block_sizeP );
	if ( *blockP == (unsigned char*) 0 )
	    {
	    (void) fprintf( stderr, "%s: out of memory\n", argv0 );
	    exit( 1 );
	    }
	*block_lenP = 0;
	}

    /* Read the block starter. */
    ich = getc( fp );
    if ( ich == EOF )
	return 0;
    (*blockP)[0] = (unsigned char) ich;
    if ( (*blockP)[0] != JPEG_BEGIN )
	return 0;
    /* Read the marker. */
    ich = getc( fp );
    if ( ich == EOF )
	return 0;
    (*blockP)[1] = (unsigned char) ich;
    /* Some block types are just two bytes without a length. */
    if ( (*blockP)[1] == JPEG_SOI || (*blockP)[1] == JPEG_EOI )
	{
	*block_lenP = 2;
	return 1;
	}
    /* Read the length high-byte. */
    ich = getc( fp );
    if ( ich == EOF )
	return 0;
    (*blockP)[2] = (unsigned char) ich;
    /* Read the length low-byte. */
    ich = getc( fp );
    if ( ich == EOF )
	return 0;
    (*blockP)[3] = (unsigned char) ich;

    /* Calculate the real lengths. */
    stoopid_len = (int) ( (*blockP)[2] ) * 256 + (int) ( (*blockP)[3] );
    data_len = stoopid_len - 2;
    *block_lenP = data_len + 4;

    /* Expand the block if necessary. */
    if ( *block_lenP > *block_sizeP )
	expand_block( blockP, block_sizeP, *block_lenP * 3 / 2 );

    /* And read the rest of the block. */
    if ( data_len > 0 )
	if ( fread( &((*blockP)[4]), data_len, 1, fp ) != 1 )
	    return 0;

    /* If it's an SOS block, read the scan data too.  There's no
    ** length field for this, we just keep reading bytes until
    ** we see a 0xFF not followed by a 0x00.  Yes, this is stupid.
    */
    got_begin = 0;
    for (;;)
	{
	ich = getc( fp );
	if ( ich == EOF )
	    return 0;
	if ( got_begin && ich != 0 )
	    {
	    /* Fixup. */
	    (void) ungetc( ich, fp );
	    (void) ungetc( JPEG_BEGIN, fp );
	    --*block_lenP;
	    break;
	    }
	if ( ich == JPEG_BEGIN )
	    got_begin = 1;
	else
	    got_begin = 0;
	if ( *block_lenP + 1 > *block_sizeP )
	    expand_block( blockP, block_sizeP, *block_lenP * 2 );
	(*blockP)[*block_lenP] = (unsigned char) ich;
	++*block_lenP;
	}

    return 1;
    }


static void
expand_block( unsigned char** blockP, size_t* block_sizeP, size_t new_size )
    {
    *block_sizeP = new_size;
    *blockP = (unsigned char*) realloc( (void*) *blockP, *block_sizeP );
    if ( *blockP == (unsigned char*) 0 )
	{
	(void) fprintf( stderr, "%s: out of memory\n", argv0 );
	exit( 1 );
	}
    }
